/*
 * Copyright (C) 2005, Edmundo Albuquerque de Souza e Silva.
 *
 * This file may be distributed under the terms of the Q Public License
 * as defined by Trolltech AS of Norway and appearing in the file
 * LICENSE.QPL included in the packaging of this file.
 *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTY OF ANY KIND, INCLUDING
 * THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "buffer.h"


Buffer::Buffer( int qtd_slots )
{
    pthread_mutex_init( &mutex, NULL );
    qtdItems       = 0;
    qtdSlots       = qtd_slots;
    lastKnownDrops = 0;

    for( int i = 0; i < qtdSlots; i++ )
    {
        freeList[ i ] = true;
        busyList[ i ] = -1;
        slot[ i ].seqNumber = 0;
        slot[ i ].applLoss  = 0;
    }
}

Buffer::~Buffer()
{
    pthread_mutex_destroy( &mutex );
}

int Buffer::insert( unsigned long seqNumber, unsigned char *data,
                    unsigned long dataSize, char flags )
{
    int i, j;
    int status;

    status = 0;
    pthread_mutex_lock( &mutex );

    if( qtdItems < qtdSlots )
    {
        i = 0;
        status = 1;

        while( (i < qtdSlots) && (busyList[ i ] != -1) )
        {
            if( slot[ busyList[ i ] ].seqNumber > seqNumber )
            {
                status = 1;
                break;
            }
            if( slot[ busyList[ i ] ].seqNumber == seqNumber )
            {
                status = -1;
                break;
            }
            i++;
        }

        if( status == 1 )
        {
            /* make room for the new slot */
            if( busyList[ i ] != -1 )
            {
                for( j = qtdItems - 1; j >= i; j-- )
                    busyList[ j + 1 ] = busyList[ j ];
            }

            /* Get the first available slot and set it to busy */
            for( j = 0; j < qtdSlots; j++ )
                if( freeList[ j ] == true )
                {
                    busyList[ i ] = j;
                    freeList[ j ] = false;
                    break;
                }

            /* Put the data in there */
            slot[ j ].seqNumber = seqNumber;
            slot[ j ].dataSize  = dataSize;
            slot[ j ].flags     = flags;
            slot[ j ].applLoss  = 0;
            memcpy( slot[ j ].data, data, dataSize );
            qtdItems++;
        }
    }
    else
    {
        slot[ busyList[ qtdItems - 1 ] ].applLoss++;

//        fprintf( stderr, "[buff insert] Dropei o pacote %u\n", slot[ j ].seqNumber + 1 );
    }

    pthread_mutex_unlock( &mutex );

    return( status );
}

int Buffer::remove( unsigned long *seqNumber, unsigned char *data,
                    unsigned long *dataSize, char *flags, unsigned long *drops )
{
    int status;

    status = 0;
    pthread_mutex_lock( &mutex );
    if( qtdItems > 0 )
    {
        *seqNumber = slot[ busyList[ 0 ] ].seqNumber;
        *dataSize  = slot[ busyList[ 0 ] ].dataSize;
        *flags     = slot[ busyList[ 0 ] ].flags;
        memcpy( data, slot[ busyList[ 0 ] ].data, *dataSize );
        if( drops != NULL )
            *drops = lastKnownDrops;
        lastKnownDrops = slot[ busyList[ 0 ] ].applLoss;

        freeList[ busyList[ 0 ] ] = true;
        for( int i = 1; i < qtdItems; i++ )
            busyList[ i - 1 ] = busyList[ i ];
        qtdItems--;
        busyList[ qtdItems ] = -1;
        status = 1;
    }
    pthread_mutex_unlock( &mutex );

    return( status );
}

void Buffer::getItem( unsigned long *seqNumber, unsigned char *data,
                      unsigned long *dataSize, char *flags )
{
    pthread_mutex_lock( &mutex );
    if( qtdItems > 0 )
    {
        *seqNumber = slot[ busyList[ 0 ] ].seqNumber;
        *dataSize  = slot[ busyList[ 0 ] ].dataSize;
        *flags     = slot[ busyList[ 0 ] ].flags;
        memcpy( data, slot[ busyList[ 0 ] ].data, *dataSize );
    }
    pthread_mutex_unlock( &mutex );
}

int Buffer::getQtdItems( void )
{
    return( qtdItems );
}

unsigned long Buffer::getFirstSeqNumber( void )
{
    int status;

    status = 0;
    if( qtdItems > 0 )
        status = slot[ busyList[ 0 ] ].seqNumber;
    return( status );
}
