/*
 * Copyright (C) 2005, Edmundo Albuquerque de Souza e Silva.
 *
 * This file may be distributed under the terms of the Q Public License
 * as defined by Trolltech AS of Norway and appearing in the file
 * LICENSE.QPL included in the packaging of this file.
 *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTY OF ANY KIND, INCLUDING
 * THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include "parser.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <arpa/inet.h>

extern FILE *yyin;
extern int yyparse( void );

struct parameter *config;
int n_items = 0;

/********************************************************************************
 *                                                                              *
 * The insert_variable() inicializes each variable of the configuration file.   *
 * variable_name is put in a structure along with it's type and it's            *
 * default_value.                                                               *
 *                                                                              *
 * The type's are:                                                              *
 *     - TYPE_BOOLEAN or 1: the default_value is passed as a "string" ( true or *
 *                          false ),                                            *
 *                                                                              *
 *                                                                              *
 *     - TYPE_INTEGER or 2: the default_value is passed as an integer,          *
 *                                                                              *
 *     - TYPE_STRING  or 3: the default_value is passed as a "string",          *
 *                                                                              *
 *     - TYPE_IP_ADDRESS or 4: the default_value is passed as a "string" ( the  *
 *                             number and dot notation should be used here.     *
 *                             ex: "120.122.134.143" ),                         *
 *                                                                              *
 *     - TYPE_REDUN or 5: the default_value is passed as a "string" ( it        *
 *                          should be used as: integer colon integer or         *
 *                          int : int :: int : int.                             *
 *                          ex: "1:2" or "1:2::3:6" ).                          *
 *                                                                              *
 *******************************************************************************/
void insert_variable( char *variable_name, int type, void * default_value )
{
    n_items++;
    
    config = realloc( config, n_items*sizeof( struct parameter) );
    strcpy( config[n_items-1].variable, variable_name );
    config[n_items-1].type = type;
    
    if( default_value != NULL )
    {
        switch( type )
        {
            case TYPE_BOOLEAN:   if( strcmp( (char *)default_value,"true" ) == 0 )                                
                                     config[n_items-1].v.bool_value = '1';
                                 else
                                     if( strcmp( (char *)default_value,"false" ) == 0 )                                   
                                         config[n_items-1].v.bool_value = '0';
                                 break;
                                  
            case TYPE_INTEGER:   config[n_items-1].v.int_value = (int) default_value;
                                 break;
                                  
            case TYPE_REDUN:
            case TYPE_ID:
            case TYPE_STRING:     config[n_items-1].v.str_value = strdup( 
                                                      ( char *) default_value );
                                  break;
                                  
            case TYPE_IP_ADDRESS: config[n_items-1].v.ip_value.sin_addr.s_addr = inet_addr( (char *)default_value );
                                  break;
        }
        config[n_items-1].used = 1;  
    }
    else
        config[n_items-1].used = 0;
}

/********************************************************************************
 *                                                                              *
 * The parse_config() calls the yyparse() function to start parsing the         *
 * file_name file.                                                              *
 *                                                                              *
 *******************************************************************************/
int parse_config( char * file_name )
{
    yyin = fopen( file_name, "r" );
    if( !yyin )
    {
        fprintf( stderr, "Error while opening the configuration file.\n" );
        return -1;
    }
    
    else if( yyparse( ) != 0 )
    {
         fprintf( stderr, "Error while parsing the configuration file.\n" );
         return -1;
    }
         
    return 1;
}

/********************************************************************************
 *                                                                              *
 * The get_value() puts the value of variable_name in value.                    *
 *                                                                              *
 *******************************************************************************/
int get_value( char *variable_name, void *value )
{
    int i;  
    
    for( i=0; i<n_items; i++ )
    {
        if( ( strcmp( config[i].variable, variable_name ) == 0 ) && ( config[i].used == 1 ) )
        {
            switch( config[i].type )
            {
                case TYPE_BOOLEAN:    *((char *)value) = config[i].v.bool_value;
                                      break;
                                      
                case TYPE_INTEGER:    *((int *)value) = config[i].v.int_value;
                                      break; 
                                      
                case TYPE_REDUN:
                case TYPE_ID:
                case TYPE_STRING:     strcpy( value, config[i].v.str_value ); 
                                      break;
                                      
                case TYPE_IP_ADDRESS: strcpy( value, inet_ntoa( config[i].v.ip_value.sin_addr ) );
                                      break;
            }
            
            return 1;
        } 
    }
    
    return 0;
}

int yywrap( void )
{
    return 1;
}
