/*
 * Copyright (C) 2005, Edmundo Albuquerque de Souza e Silva.
 *
 * This file may be distributed under the terms of the Q Public License
 * as defined by Trolltech AS of Norway and appearing in the file
 * LICENSE.QPL included in the packaging of this file.
 *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTY OF ANY KIND, INCLUDING
 * THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <netinet/in.h>
#include <time.h>

#include "vvCodecs.h"
#include "globals.h"
#include "control.h"
#include "net.h"
#include "trace.h"
#include "version.h"
#include "vvsignal.h"
#include "stats.h"

extern USERINFO vvserverInfo;
extern char generateTrace;
extern bool fm_running;

static void processStartPckt( PCONNECTINFO, StartPacket *, struct sockaddr_in, int);
static void processEndPckt( PCONNECTINFO, struct sockaddr_in );
static int fillStatPacket( PCONNECTINFO, struct sockaddr_in, StatPacket * );

void printMessage( char *, ... );
int cmpIP( char *, char * );
void getIP( char *, char * );
int numberOfConnections( PCONNECTINFO );

void updateConnectionStatus( PCONNECTINFO, struct sockaddr_in );


/**
 * Send a control packet of type type to dest. 
 * This routine fullfills the control packet with the appropriate data 
 * according to type.
 * @return number of bytes sent.
 */
int sendControlPckt( PCONNECTINFO info, char type, struct sockaddr_in dest,
                     int length )
{
    ControlPacket * packet = NULL;
    int bysend = 0;

    if( type == VVControlPacket::StartPacket )
    {
        StartPacket vvPacket;
        vvPacket.type    = type;
        vvPacket.version = VERSION;

        strcpy( vvPacket.username, vvserverInfo.username );
        strcpy( vvPacket.hostname, vvserverInfo.hostname );

        bysend = sizeof( StartPacket );
        packet = &vvPacket;	
    }
    else if( type == VVControlPacket::EndPacket )
    {
        EndPacket vvPacket;
        vvPacket.type    = type;
        vvPacket.version = VERSION;

        bysend = sizeof( EndPacket );
        packet = &vvPacket;
    }
    else if( type == VVControlPacket::BusyPacket )
    {
         BusyPacket vvPacket;
         vvPacket.type    = type;
         vvPacket.version = VERSION;

         bysend = sizeof( BusyPacket );
         packet = &vvPacket;
    }
    else if( type == VVControlPacket::TalkPacket )
    {
         TalkPacket vvPacket;
         vvPacket.type    = type;
         vvPacket.version = VERSION;
#ifdef WALKIETALKIE
         bysend = sizeof( TalkPacket );
#endif
         packet = &vvPacket;
    }		
    else if( type == VVControlPacket::SilPacket )
    {
         SilPacket vvPacket;
         vvPacket.type    = type;
         vvPacket.version = VERSION;

//TODO:  vvPacket.flags   = silenceBit;
         bysend = sizeof( SilPacket );
         packet = &vvPacket;
    }
    else if( type == VVControlPacket::UserPacket )
    {
         UserPacket vvPacket;
         vvPacket.type    = type;
         vvPacket.version = VERSION;

         vvPacket.codec = vvserverInfo.codec;
         vvPacket.redun = vvserverInfo.redun;
         vvPacket.silenceUser = vvserverInfo.silenceUser;

         bysend = sizeof( UserPacket );
         packet = &vvPacket;
    }
    else if( type == VVControlPacket::StatPacket )
    {
        StatPacket vvPacket;
        vvPacket.type    = type;
        vvPacket.version = VERSION;

        fillStatPacket( info, dest, &vvPacket );

        bysend = sizeof( StatPacket );
        packet = &vvPacket;
    }
    else if( type == VVControlPacket::StatusPacket )
    {
        StatusPacket vvPacket;
        vvPacket.type    = type;
        vvPacket.version = VERSION;

        bysend = sizeof( StatusPacket );
        packet = &vvPacket;
    }
    else
    {
        bysend = 0;
    }

    return( sendto( info->socCtrl, (char *)packet, bysend, 0,
            (struct sockaddr *)&dest, length ) );
}

/**
 * Send a control packet of type type to FreeMeeting. 
 * @return number of bytes sent.
 */
int sendFMControlPckt( PCONNECTINFO info, FMControlPacket * packet,
                       int length, int type )
{
    // FreeMeeting Packets
    packet->type    = type;
    packet->version = FM_VERSION;
    packet->tool    = VVServer;
    
    return( sendto( info->socCtrl, (char *)packet, length, 0,
            (struct sockaddr *)&info->FMdest, sizeof( struct sockaddr_in ) ) );
}

/**
 * This thread waits for a control packet and process it.
 */
void *recvControlPckt( void *arg )
{
    int length;
    int byread;
    PCONNECTINFO info;
    VVControlPacket * packet;
    char  *buffer;
    struct sockaddr_in from;
    int maxPacket;

    // Buffer has to have the maximum packet size
    maxPacket = sizeof( StartPacket );
    info = (PCONNECTINFO)arg;
    while( 1 )
    {
        length = sizeof( struct sockaddr_in );
        buffer = (char *)malloc( maxPacket );
        bzero( buffer, maxPacket );
        byread = recvfrom( info->socCtrl, buffer, maxPacket, 0,
                           (struct sockaddr *)&from, (socklen_t *)&length );

        if( byread <= 0 )
            continue;
 
        packet = ( VVControlPacket * )buffer;

        /* Checking protocol version */
        /* First checks if it is a FMPacket or a VVPacket */
        if( packet->type < FMControlPacket::FMConnectPacket )
        {
            if( packet->version != VERSION )
            {
                if( fm_running )
                {
                    fprintf( stderr, "This VVServer is not compatible with Vivavoz." );
                }
                continue;
            }
        }
        else /* FMControlPacket */
        {
            if( packet->version != FM_VERSION )
            {
                /* this VVServer is not compatible with FreeMeeting,
                   exiting */
                if( fm_running )
                {
                    fprintf( stderr, "This VVServer is not compatible with FreeMeeting, exiting." );
                    exit( 1 );
                 }
                else /* ignore packet */
                    continue;
            }
        }
       
        switch( packet->type )
        {
            case VVControlPacket::StartPacket:
                if( info->type != TYPE_MULTICAST )
                   processStartPckt( info, (StartPacket *)packet,
                                     from, length );
                break;

    	    case VVControlPacket::EndPacket:
                processEndPckt( info, from );
                break;

    	    case VVControlPacket::StatusPacket:           
                updateConnectionStatus( info, from );
                break;
            /* Do nothing */
	        case VVControlPacket::StatPacket:

    	    case VVControlPacket::SilPacket:

	        case VVControlPacket::UserPacket:
                break;

            // FreeMeeting Packets
   	        case FMControlPacket::FMConnectPacket:

            case FMControlPacket::FMStatisticsPacket:
                break;

            case FMControlPacket::FMDisconnectPacket:
                endConnections( info );
                break;

        }
    }

    return( NULL );
}

void endConnections( PCONNECTINFO info )
{
    PDESTINATION cur;
    struct sockaddr_in dstAddr;
    char ip[15];

    cur = info->dests;

    printMessage( "Ending vvserver connections...\n" );
    
    /* End vvserver connections */
    while( cur != NULL )
    {
        getIP( (char *)&cur->addr, ip );
        printMessage( "Ending connection with %s\n", ip );

        memcpy( (char *)&dstAddr, (char *)&cur->addr,
                sizeof(struct sockaddr_in) );
        dstAddr.sin_port = htons( DEFAULT_PORT );
        sendControlPckt( info, VVControlPacket::EndPacket, dstAddr,
                         sizeof( struct sockaddr_in ) );

        if( generateTrace )
            traceClose( cur );
        
        if( fm_running )
        {
            FMDisconnectPacket packet;
 
            strcpy( packet.username, cur->username );
            strcpy( packet.hostname, cur->hostname );

            sendFMControlPckt( info, &packet,
                               sizeof( FMDisconnectPacket ),
                               FMControlPacket::FMDisconnectPacket );
        }


        info->dests = cur->next;

        delete cur->codecr;
        delete cur->codecs;
        free( cur );

        cur = info->dests;
    }

    info->connected = NOT_CONNECTED;
    alarmOff(); 
}

/**
 * Process START_PACKETs.
 * It's a three way hand shake.
 * - The client that wants to connect to vvserver send a START_PACKET to
 *   Vvserver.
 * - Vvserver adds this client to the connection's list and send a START_PACKET
 *   to the client. Vvserver changes his state to SYN_RECV.
 * - The client sends another START_PACKET to Vvserver and establishes the
 *   connection.
 * - When Vvserver receives this START_PACKET it changes its state to
 *   ESTABLISHED.
 */
static void processStartPckt( PCONNECTINFO info, StartPacket * start,
                              struct sockaddr_in from, int length )
{
    PDESTINATION dest, prev;
    char ip[15];

    prev = dest = info->dests;

    /* Searching the new peer into the connection list 
       if not in list -> dest = NULL */
    while( dest != NULL )
    {
        if( cmpIP( (char *)&dest->addr, (char *)&from ) == 0 )
            break;

        prev = dest;
        dest = dest->next;
    }

    /* Inserting the new peer into the connection list */
    if( dest == NULL )
    {
        dest = (PDESTINATION)malloc( sizeof( DESTINATION ) );
        bzero( dest, sizeof( DESTINATION ) );

        dest->state = NOT_CONNECTED;
        memcpy( (char *)&dest->addr, (char *)&from, 8 );
        dest->addr.sin_port = htons( DEFAULT_PORT + 1 );
        dest->codecr = new vvCodecs( MODE_DECODER );
        dest->codecs = new vvCodecs( MODE_ENCODER );
        strcpy( dest->hostname, start->hostname );
        strcpy( dest->username, start->username );

        if( prev != NULL )
            prev->next  = dest;
        else
            info->dests = dest;

        dest->next = NULL;
    }

    if( dest->state != ESTABLISHED )
    {
        switch( dest->state )
        {
            case NOT_CONNECTED:
                dest->state = SYN_RECV;
                sendControlPckt( info, VVControlPacket::StartPacket, from,
                                 length );
                break;

            case SYN_SENT:
                sendControlPckt( info, VVControlPacket::StartPacket,
                                 from, length );

            /* Fall through */
            case SYN_RECV:
                info->type  = TYPE_UNICAST;
                initStats( dest );

                /* initialize redundancy reception routine */
                dest->recvRedun = redun_CreateRecvChain( 24, 2048 );
                dest->buffer = new Buffer( 25 );

                /* Initializing redundancy sending routine */
                dest->sendRedun = NULL;

                dest->isSilent = TRUE;
                dest->lastStatusPckt = -1;

                if( generateTrace )
                {
                    initializeTraceInfo( dest, start->hostname );
                    traceOpen( dest, info->logDir, start->hostname );
                }

                if( numberOfConnections( info ) == 1 )
                {
                    alarmOn();
                    info->connected = ESTABLISHED; 
                }

                getIP( (char *)&from, ip );
                printMessage( "%s has connected to vvserver", ip );

                if( fm_running )
                {
                    FMAckPacket ackPacket;

                    strcpy( ackPacket.username, start->username );
                    strcpy( ackPacket.hostname, start->hostname );

                    sendFMControlPckt( info, &ackPacket, sizeof( FMAckPacket ),
                                    FMControlPacket::FMAckPacket );
                }
                
                break;

            case ESTABLISHED:
                if( dest->state == TYPE_UNICAST )
                    sendControlPckt( info, VVControlPacket::BusyPacket,
                                     from, length );
                break;
        }
    }
    else
    {
        if( info->type == TYPE_UNICAST )
        {
            #ifdef _DEBUG
            fprintf( stderr, "\tPeer is busy. Ooops\n" );
            #endif
            sendControlPckt( info, VVControlPacket::BusyPacket, from, length );
        }
    }
}

/**
 * Process END_PACKETs. 
 * Removes the client "from" from the connection's list.
 */
static void processEndPckt( PCONNECTINFO info, struct sockaddr_in from )
{
    PDESTINATION aux, prev;
    char ip[15];

    prev = NULL;
    aux  = info->dests;
    while( aux != NULL )
    {
        if( cmpIP( (char *)&aux->addr, (char *)&from ) == 0 )
        {
            if( generateTrace )
                traceClose( aux );

            if( fm_running )
            {
                FMDisconnectPacket packet;

                strcpy( packet.username, aux->username );
                strcpy( packet.hostname, aux->hostname );

                sendFMControlPckt( info, &packet,
                                   sizeof( FMDisconnectPacket ),
                                   FMControlPacket::FMDisconnectPacket );
            }

            if( prev == NULL )
                info->dests = aux->next;
            else
                prev->next  = aux->next;

            delete aux->codecr;
            delete aux->codecs;
            free( aux );

            getIP( (char *)&from, ip );
            printMessage( "%s has ended connection with vvserver", ip );

            break;
        }
        prev = aux;
        aux  = aux->next;
    }

    if( numberOfConnections( info ) == 0 )
    {
        info->connected = NOT_CONNECTED;
        alarmOff();
    }

    #ifdef _DEBUG
    printConnectionList( info );
    #endif
}

/**
 * Processes START_PACKETs.
 * Connection's steps description: (Three way hand shake)
 * - The client that wants to connect to vvserver send a START_PACKET to
 *   Vvserver.
 * - Vvserver adds this client to the connection's list and send a START_PACKET
 *   to the client. Vvserver changes his state to SYN_RECV.
 * - The client sends another START_PACKET to Vvserver and establishes the
 *   connection.
 * - When Vvserver receives this START_PACKET it changes its state to
 *   ESTABLISHED.
 */
static int fillStatPacket( PCONNECTINFO info, struct sockaddr_in from,
                           StatPacket * packet )
{
    int status;
    PDESTINATION dst;

    status = 0;
    dst = info->dests;
    while( dst != NULL )
    {
        if( cmpIP( (char *)&dst->addr, (char *)&from ) == 0 )
            break;
        dst = dst->next;
    }

    if( dst != NULL )
    {
        packet->stats.sent       = htonl( dst->packets.sent );
        packet->stats.lost       = htonl( dst->packets.lost );
        packet->stats.received   = htonl( dst->packets.received );
        packet->stats.recovered  = htonl( dst->packets.recovered );
        packet->stats.dropped    = htonl( dst->packets.dropped );
        packet->stats.outoforder = htonl( dst->packets.outoforder );
        packet->stats.notplayed  = htonl( dst->packets.notplayed );
        packet->stats.successMean.tlong = htonl( dst->packets.successMean.tlong );
        packet->stats.lostMean.tlong    = htonl( dst->packets.lostMean.tlong );
        status = 1;
    }

    return( status );
}

int numberOfConnections( PCONNECTINFO info )
{
    PDESTINATION aux;
    int num = 0;

    aux = info->dests;

    while( aux != NULL )
    {
        num++;
        aux = aux->next;
    }

    return ( num );
}

void updateConnectionStatus( PCONNECTINFO info, struct sockaddr_in from )
{
    PDESTINATION cur;

    cur = info->dests;
    while( cur != NULL )
    {
        if( cmpIP( (char *)&cur->addr, (char *)&from ) == 0 )
        {
            /* it's the first status packet received */
            if( cur->lastStatusPckt == -1 )
                cur->state = ESTABLISHED;

            time( &cur->lastStatusPckt );
            break;
        }

        cur = cur->next;
    }
}
