/*
 * Copyright (C) 2005, Edmundo Albuquerque de Souza e Silva.
 *
 * This file may be distributed under the terms of the Q Public License
 * as defined by Trolltech AS of Norway and appearing in the file
 * LICENSE.QPL included in the packaging of this file.
 *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTY OF ANY KIND, INCLUDING
 * THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
#include <sys/wait.h>
#include <execinfo.h>
 
#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>
#include <stdarg.h>
#include <signal.h>
#include <errno.h>

#include "vvCodecs.h"
#include "globals.h"
#include "net.h"
#include "control.h"
#include "forward.h"
#include "trace.h"
#include "buffer.h"
#include "version.h"

extern "C"
{
#include "confparser/parser.h"
#include "gsm.h"
};

#define NO_REDUN_SCHEMES 2

#define CODEC          0
#define REDUN          1
#define GENERATE_TRACE 2
#define TRACE_FILE_DIR 3

#define DEFAULT_CONF_FILE INSTALL_DIR "/etc/vvserver.conf"

char        generateTrace;
pthread_t   tWSP;

USERINFO    vvserverInfo;
CONNECTINFO conInfo;

int daemon_flag;
char * vvConfFilePath;

bool fm_running;

void printMessage( char * msg, ... )
{
    va_list params;
    char fmtMsg[ 256 ];

    va_start( params, msg );
    vsprintf( fmtMsg, msg, params );
    va_end( params );

    if( daemon_flag )
    {
        syslog( LOG_INFO, fmtMsg );
    }    
    else
    {
        fprintf( stderr, "%s\n", fmtMsg );
    }
}

/**
 * Compares two given IP addresses
 */
int cmpIP( char *ip1, char *ip2 )
{
    char *ptr1, *ptr2;

    ptr1 = ip1;
    ptr2 = ip2;
    ptr1 += 4;
    ptr2 += 4;

    return( memcmp( ptr1, ptr2, 4 ) );
}

void getIP( char *ip, char * str )
{
    unsigned char *ptr;

    ptr = (unsigned char *)ip;
    ptr += 4;
    
    sprintf( str, "%d.%d.%d.%d", *ptr, *(ptr+1), *(ptr+2), *(ptr+3) );
}

#ifdef _DEBUG
void printConnectionList( PCONNECTINFO info )
{
    PDESTINATION aux;
    int i=1;

    aux = info->dests;
    
    fprintf( stderr, "Printing list of connected clients:\n" );
    while( aux != NULL )
    {
        char ip[ 15 ];
        fprintf( stderr, "Client %d: ", i );
        getIP( (char *)&aux->addr, ip );
        fprintf( stderr, "%s\n", ip );        
        i++;
        
        aux = aux->next;
    }
}
#endif

/**
 * Reads the configuration file. Calls the parser in ./confparser.
 * @return 1 - upon failure
 *         0 - upon success
 */
int initializeConfigOptions()
{
    insert_variable( "CODEC", TYPE_ID, ( void * ) "PCM" );
    insert_variable( "REDUN", TYPE_REDUN, (void *) "1:2::3:6" );
    insert_variable( "GENERATE_TRACE", TYPE_BOOLEAN, (void *) "false" );
    insert_variable( "TRACE_FILE_DIR", TYPE_STRING, (void *) "/tmp/" );

    if( parse_config( vvConfFilePath ) == -1 )
        return 1;
        
    return 0;
}

/**
 * Initializes and updates the vvserverInfo structure with the values 
 * read in the configuration file.
 * @return 1 - upon failure
 *         0 - upon success
 */
int initializeUserInfo()
{
    unsigned char codec;
    int redun, daemon;
    int status = 0;

    /* Codec */
    if( strcasecmp( config[CODEC].v.str_value, "gsm" ) == 0 )
        codec = CODEC_GSM;
    else if( strcasecmp( config[CODEC].v.str_value, "pcm" ) == 0 )
        codec = CODEC_NONE;
    else if( strcasecmp( config[CODEC].v.str_value, "speex_0" ) == 0 )
        codec = CODEC_SPEEX | CODEC_SPEEX_0;
    else if( strcasecmp( config[CODEC].v.str_value, "speex_1" ) == 0 )
        codec = CODEC_SPEEX | CODEC_SPEEX_1;
    else if( strcasecmp( config[CODEC].v.str_value, "speex_2" ) == 0 )
        codec = CODEC_SPEEX | CODEC_SPEEX_2;
    else if( strcasecmp( config[CODEC].v.str_value, "speex_3" ) == 0 )
        codec = CODEC_SPEEX | CODEC_SPEEX_3;
    else if( strcasecmp( config[CODEC].v.str_value, "speex_5" ) == 0 )
        codec = CODEC_SPEEX | CODEC_SPEEX_5;
    else if( strcasecmp( config[CODEC].v.str_value, "speex_7" ) == 0 )
        codec = CODEC_SPEEX | CODEC_SPEEX_7;
    else if( strcasecmp( config[CODEC].v.str_value, "speex_9" ) == 0 )
        codec = CODEC_SPEEX | CODEC_SPEEX_9;
    else if( strcasecmp( config[CODEC].v.str_value, "speex_10" ) == 0 )
        codec = CODEC_SPEEX | CODEC_SPEEX_10;
    else
    {
        printMessage( "Configuration File: Invalid Codec - %s",
                      config[CODEC].v.str_value );
        codec = vvserverInfo.codec;
        status = 1;
    }
    
    /* Redun */
    if( strcasecmp( config[REDUN].v.str_value, "none" ) == 0 )
        redun = 0;
    else if( strcasecmp( config[REDUN].v.str_value, "1:2" ) == 0 )
        redun = 1;
    else if( strcasecmp( config[REDUN].v.str_value, "1:2::3:6" ) == 0 )
        redun = 2;
    else
    {
        printMessage( "Configuration File: Invalid Redun - %s",
                      config[REDUN].v.str_value );
        redun = vvserverInfo.redun;
        status = 1;
    }

    /* Trace */
    if( config[GENERATE_TRACE].v.bool_value == '1' )
        generateTrace = 1;
    else if( config[GENERATE_TRACE].v.bool_value == '0' )
        generateTrace = 0;
    else
    {
        printMessage( "Configuration File: Invalid Generate Trace Option - %c",
                      config[GENERATE_TRACE].v.bool_value );
        status = 1;
    }

    /* Trace File Directory */
    if( generateTrace )
    {
        if( access( config[TRACE_FILE_DIR].v.str_value, F_OK | W_OK ) < 0 )
        {
            printMessage( "%s: %s", config[TRACE_FILE_DIR].v.str_value, 
                          strerror( errno )  );
            status = 1;
        }
        else
            strcpy( conInfo.logDir, config[TRACE_FILE_DIR].v.str_value );
    }

    daemon = daemon_flag;

    bzero( (char *)&vvserverInfo, sizeof(vvserverInfo) );

    vvserverInfo.format = 0;
    vvserverInfo.codec  = codec;
    vvserverInfo.redun  = redun;

    strcpy( vvserverInfo.username, "vvserver " );
    strcat( vvserverInfo.username, STR_VERSION );

    if( gethostname( vvserverInfo.hostname, 80 ) != 0 )
        strcpy( vvserverInfo.hostname, "unknown" );

    return( status );
}

int createSocket( int port )
{
    int soc;
    struct sockaddr_in saddr;

    soc = socket( PF_INET, SOCK_DGRAM, 0 );

    saddr.sin_family      = AF_INET;
    saddr.sin_port        = htons( port );
    saddr.sin_addr.s_addr = INADDR_ANY;
    if( bind( soc, (struct sockaddr *)&saddr, sizeof( saddr )) < 0 )
    {
        printMessage( "bind: %s", strerror( errno ) );
        exit( -1 );
    }

    return( soc );
}

/**
 * initializes the connection variables and creates the recvControlPckt
 * thread
 */
void initializeNet( PCONNECTINFO conInfo )
{
    conInfo->connected       = NOT_CONNECTED;
    conInfo->type            = TYPE_DISCONNECTED;
    conInfo->dests           = NULL;
    conInfo->socCtrl = createSocket( DEFAULT_PORT );
    conInfo->socData = createSocket( DEFAULT_PORT + 1 );

    if( fm_running )
    {
        bzero( (char *)&conInfo->FMdest, sizeof( conInfo->FMdest ) );
        conInfo->FMdest.sin_family = AF_INET;
        conInfo->FMdest.sin_port  = htons( FREEMEETING_PORT );
    }

    if( pthread_create( &tWSP, NULL, recvControlPckt, (void *)conInfo ) != 0 )
        exit( -1 );
}

/**
 * turns vvserver into a daemon
 */
void createDaemon()
{
    pid_t pid;

    /* Now it is time to turn vvserver into a daemon */
    if( ( pid = fork() ) < 0 )
    {
        printMessage( "Unable to fork - exiting program" );
        exit( 0 );
    }
    else
    {
        if( pid > 0 )
            exit( 0 );

        /* child */
        setsid();       /* becomes session leader */
        chdir( "/" );   /* change directory */
    }
}

/**
 * Handles the HUP(-1) signal. Calls the vvserver update configuration
 * functions
 */
void sighupHandler( int sigNum )
{
    printMessage( "Updating vvserver configuration ..." );
    initializeConfigOptions();
    initializeUserInfo();
}

/**
 * Handles the TERM(-15) and INT(ctrl+c) signal. Closes all vvserver
 * connections and exits the program
 */
void sigtermHandler( int sigNum )
{
    endConnections( &conInfo );
    
    pthread_cancel( tWSP );
    pthread_join( tWSP, NULL );

    close( conInfo.socData );

    printMessage( "vvserver ended" );

    exit( 0 );
}

/**
 * Handles the SEGFAULT.
 */
void show_stackframe()
{
    void *trace[16];
    char **messages = (char **)NULL;
    int i, trace_size = 0;

    trace_size = backtrace( trace, 16 );
    messages = backtrace_symbols( trace, trace_size );
    fprintf( stderr, "[bt] Execution path:\n");
    for( i = 0; i < trace_size; ++i )
        fprintf( stderr, "[bt] %s\n", messages[i]);
}

void segfaultHandler( int sig )
{
    fprintf( stderr, "SEGMENTATION FAULT!\n" );
    show_stackframe();
    
    _exit( 1 );
}

/**
 * Checks if vvserver's configuration file is at env_variable/path.
 * If env_variable is NULL checks only in path.
 * @return 0 - upon error
 *         1 - upon success
 */
int checkConfPath( const char * env_variable, char * path )
{
    if( env_variable != NULL )
    {
        if( getenv( env_variable ) != NULL )
        {
            vvConfFilePath = (char *) malloc( strlen( getenv( env_variable ) )
                                              + strlen( path ) + 1 );
            strcpy( vvConfFilePath, getenv( env_variable ) );
        }
        else
        {
            printMessage( "%s environment variable does not match.\n",
                          env_variable );
            return 0;
        }
    }
    else
        vvConfFilePath = (char *) malloc( strlen( path ) + 1 );

    strcat( vvConfFilePath, path );

    printMessage( "Searching %s ...", vvConfFilePath );

    if( access( vvConfFilePath, F_OK | R_OK ) < 0 )
    {
        return 0;
    }

    return 1;
}

void printOptions()
{
    printf( "Usage: vvserver [OPTION]\n\n" );
    printf( "Options:\n" );
    printf( "\t-d\t\tstarts as daemon\n" );
    printf( "\t-m\t\tstarts with FreeMeeting\n" );
    printf( "\t-v\t\tprints version\n" );
    printf( "\t-f PATH\t\treads the configuration file from PATH\n" );
    printf( "\t\t\tDefault paths are: $HOME\n" );    
    printf( "\t\t\t                   /etc\n" );            
}

int main( int argc, char **argv )
{
    const char *optstring = "mvdhf:";
    int c;
    char ident[30];

    /* Handling Signals */
    signal( SIGHUP, sighupHandler );
    signal( SIGTERM, sigtermHandler );
    signal( SIGINT, sigtermHandler );
    signal( SIGSEGV, segfaultHandler );

    daemon_flag  = 0;

    vvConfFilePath = NULL;

    strcpy( ident, "vvserver-" );
    strcat( ident, STR_VERSION );

    openlog( ident, LOG_PID, LOG_DAEMON );

    bzero( (char *)&vvserverInfo, sizeof( vvserverInfo ) );

    while( ( c = getopt( argc, argv, optstring ) ) != -1 )
    {
        switch( c )
        {
            case 'h':
                printOptions();
                return 0;
            case 'd':
                daemon_flag = 1;
                createDaemon();
                break;
            case 'f':
                vvConfFilePath = (char *) malloc( strlen( optarg ) +
                                 strlen( "/vvserver.conf" ) + 1 );
                strcpy( vvConfFilePath, optarg );
                if( optarg[ strlen( optarg ) - 1 ] != '/' )
                    strcat( vvConfFilePath, "/" );
                strcat( vvConfFilePath, "vvserver.conf" );
                break;
            case 'm':
                fm_running = true;
                break;
            case 'v':
                printf( "VVServer %s - Vivavoz Server.\n\n", STR_VERSION );
                return 0;
             default:
                exit( 1 );
        }
    }

    printMessage( "vvserver starting ..." );

    /* if has a alternative path */
    if( vvConfFilePath != NULL )
    {
        if( access( vvConfFilePath, F_OK | R_OK ) < 0 )
        {
            printMessage( "%s: %s", vvConfFilePath, strerror( errno ) );
            return 1;
        }
    }
    /* verify if vvserver.conf's path exists: $HOME/.vvserver.conf  */
    else if( !checkConfPath( "HOME", "/.vvserver.conf" ) )
    {

        /* INSTALL_DIR/etc/vvserver.conf */
        if( !checkConfPath( NULL, DEFAULT_CONF_FILE ) )
        {
            printMessage( "Could not find \"vvserver.conf\".\nYou could specify your configuration file location running vvserver with \"-f PATH\" option." );
            return 1;
        }
    }              

    printMessage( "Loading \"%s\".", vvConfFilePath );

    if( initializeConfigOptions() != 0 )
        return 1;

    if( initializeUserInfo() != 0 )
        return 1;

    initializeNet( &conInfo );

    recvAudio( &conInfo );

    return 0;
}
