/*
 * Copyright (C) 2005, Edmundo Albuquerque de Souza e Silva.
 *
 * This file may be distributed under the terms of the Q Public License
 * as defined by Trolltech AS of Norway and appearing in the file
 * LICENSE.QPL included in the packaging of this file.
 *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTY OF ANY KIND, INCLUDING
 * THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
 
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>

#include <stdio.h>
#include <stdlib.h>

#include "vvCodecs.h"
#include "globals.h"
#include "control.h"
#include "forward.h"
#include "net.h"
#include "vvsignal.h"
#include "stats.h"

#define ALIVE 180

extern CONNECTINFO     conInfo;
extern pthread_mutex_t mutexRecv;
extern bool            fm_running;
extern USERINFO        vvserverInfo;

/*extern HMMTRACE        hmm_trace;
extern pthread_mutex_t mutexHMMTrace;*/

/**
 * Performs the periodic tasks of Vvserver.
 * - Sends audio every 20 ms if still conected.
 * - Sends status packet every 10 seconds.
 * - Checks if clients are still connected every 20 seconds.
 * - Sends user packet every 20 seconds.
 * - Sends stats packet every minute.
 */
void alrmHandle()
{
    static int count  = 0;
    PDESTINATION dst;
    struct sockaddr_in dstAddr;

    /* sending audio every 20 ms if still conected */
    if( conInfo.connected != NOT_CONNECTED )
    {
        forwardData( &conInfo );
    }

    /* Sending Stats to FreeMeeting every second */
    if( fm_running && (count % 50 == 0 ) )
    {
        FMStatisticsPacket packet;
        packet.sent = htonl( vvserverInfo.packets.sent * sizeof( DATAPCKT ) );
        packet.received = htonl( vvserverInfo.packets.received * sizeof( DATAPCKT ) );
        
        sendFMControlPckt( &conInfo, &packet, sizeof( FMStatisticsPacket ),
                           FMControlPacket::FMStatisticsPacket ); 
    }
    
    /* Send status packet every 10 seconds */
    if( ( count % 500 ) == 0 )
    {
        dst = conInfo.dests;
        while( dst != NULL )
        {
            memcpy( (char *)&dstAddr, (char *)&dst->addr, 
                    sizeof(struct sockaddr_in) );
            dstAddr.sin_port = htons( DEFAULT_PORT );
            sendControlPckt( &conInfo, VVControlPacket::StatusPacket, dstAddr,
                             sizeof( struct sockaddr_in ) );
            dst = dst->next;
        }
    }

    /* Check if clients are still connected - 20 seconds */
    if( (count % 1000) == 0 )
    {
        PDESTINATION prev;
        time_t currTime;
        int diff;

        time( &currTime );
        dst = conInfo.dests;
        prev = NULL;

        while( dst != NULL )
        {
            char ip[15];

            diff = (int) difftime( dst->lastStatusPckt, currTime );

            if( diff > ALIVE )
            {
                if( prev == NULL )
                    conInfo.dests = dst->next;
                else
                    prev->next  = dst->next;
                free( dst );
 
                getIP( (char *)&dst->addr, ip );
                printMessage( "%s has been removed from list: connection lost.",
                               ip );

                break;
             }
             
             prev = dst;
             dst = dst->next;
        }
    }

    /* Send user packet every 20 seconds */
    if( (count % 1000) == 10 )
    {
        dst = conInfo.dests;
        while( dst != NULL )
        {
            memcpy( (char *)&dstAddr, (char *)&dst->addr, 
                     sizeof(struct sockaddr_in) );
            dstAddr.sin_port = htons( DEFAULT_PORT );
            sendControlPckt( &conInfo, VVControlPacket::UserPacket, dstAddr,
                             sizeof( struct sockaddr_in ) );
            dst = dst->next;
        }
    }

    /* Send stats packet every minute */
    if( count == 3000 )
    {
        dst = conInfo.dests;
        while( dst != NULL )
        {
            memcpy( (char *)&dstAddr, (char *)&dst->addr, 
                    sizeof(struct sockaddr_in) );
            dstAddr.sin_port = htons( DEFAULT_PORT );
            sendControlPckt( &conInfo, VVControlPacket::StatPacket, dstAddr,
                             sizeof( struct sockaddr_in ) );
            dst = dst->next;
        }
        
        /* TODO: TEM Q MANDAR UM DESSES POR CONEXAO!!!! */
        /* HMM Trace */
//         pthread_mutex_lock( &mutexHMMTrace );
//         sendControlPckt( &conInfo, VVControlPacket::HMMPacket, conInfo.dest,
//                          sizeof( conInfo.dest ) );
// 
//         /* Initialize trace variables */
//         memset( hmm_trace.trace, 255, MAXTRACE );
//         hmm_trace.base   += hmm_trace.offset;
//         hmm_trace.offset = 0;
//         pthread_mutex_unlock( &mutexHMMTrace );

        count = 0;
    }
    count++;
}

void alarmOn()
{
    sigset_t set;
    struct sigaction action;
    struct itimerval t;

    sigemptyset( &set );
    sigaddset( &set, SIGALRM );

    action.sa_handler = (void(*)(int))alrmHandle;
    action.sa_mask    = set;
    action.sa_flags   = SA_RESTART;
    pthread_sigmask( SIG_UNBLOCK, &set, NULL );
    sigaction( SIGALRM, &action, NULL );

    t.it_interval.tv_sec  = 0; 
    t.it_interval.tv_usec = 20000;
    t.it_value.tv_sec     = 0;
    t.it_value.tv_usec    = 20000;

    setitimer( ITIMER_REAL, &t, NULL);
}

void alarmOff()
{
    sigset_t set;
    struct itimerval t;

    t.it_interval.tv_sec  = 0; 
    t.it_interval.tv_usec = 0;
    t.it_value.tv_sec     = 0; 
    t.it_value.tv_usec    = 0;

    setitimer( ITIMER_REAL, &t, NULL );

    sigemptyset( &set );
    sigaddset( &set, SIGALRM );

    pthread_sigmask( SIG_UNBLOCK, &set, NULL );
}
